﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Markup;
using Buddy.Coroutines;
using log4net;
using Loki.Bot;
using Loki.Bot.Pathfinding;
using Loki.Bot.v3;
using Loki.Bot.Logic.Bots.BasicGrindBot;
using Loki.Game;
using Loki.Game.GameData;
using Loki.Game.Objects;
using Loki.Utilities;

namespace FlickerStrike
{
    /// <summary> </summary>
    public class FlickerStrike : IRoutine
    {
        private static readonly ILog Log = Logger.GetLoggerInstanceForType();

        // Auto-set, you do not have to change these.
        private int _raiseZombieSlot = -1;
        private int _raiseSpectreSlot = -1;
        private int _animateWeaponSlot = -1;
        private int _animateGuardianSlot = -1;
        private int _flameblastSlot = -1;
        private int _enduringCrySlot = -1;
        private int _moltenShellSlot = -1;
        private int _bloodRageSlot = -1;
        private int _rfSlot = -1;
        private int _arcticArmourSlot = -1;
        private int _vaalDiscSlot = -1;
        private readonly List<int> _curseSlots = new List<int>();
        private int _auraSlot = -1;
        private int _totemSlot = -1;
        private int _trapSlot = -1;
        private int _mineSlot = -1;
        private int _summonSkeletonsSlot = -1;
        private int _summonRagingSpiritSlot = -1;
        private int _coldSnapSlot = -1;

        private int _currentLeashRange = -1;

        private readonly Stopwatch _trapStopwatch = Stopwatch.StartNew();
        private readonly Stopwatch _totemStopwatch = Stopwatch.StartNew();
        private readonly Stopwatch _flickerstrikeStopwatch = Stopwatch.StartNew();
        private readonly Stopwatch _mineStopwatch = Stopwatch.StartNew();
        private readonly Stopwatch _animateWeaponStopwatch = Stopwatch.StartNew();
        private readonly Stopwatch _animateGuardianStopwatch = Stopwatch.StartNew();
        private readonly Stopwatch _moltenShellStopwatch = Stopwatch.StartNew();
        private readonly List<int> _ignoreAnimatedItems = new List<int>();

        private int _summonSkeletonCount;
        private readonly Stopwatch _summonSkeletonsStopwatch = Stopwatch.StartNew();

        private int _summonRagingSpiritCount;
        private readonly Stopwatch _summonRagingSpiritStopwatch = Stopwatch.StartNew();

        private bool _castingFlameblast;
        private int _lastFlameblastCharges;
        private bool _needsUpdate;

        private readonly Targeting _combatTargeting = new Targeting();

        private Targeting CombatTargeting
        {
            get { return _combatTargeting; }
        }

        // Do not implement a ctor and do stuff in it.

        #region Targeting

        private void CombatTargetingOnWeightCalculation(NetworkObject entity, ref float weight)
        {
            weight -= entity.Distance/2;

            var m = entity as Monster;
            if (m == null)
                return;
				
			// Place higher weight on adds when Kole is nearby. This should force the bot to kill the normal mobs around him first to prevent dying from them casting corpse explode on Kole.   			
			var Kole = LokiPoe.ObjectManager.GetObjectByName<Monster>("Kole");
			var myPos = LokiPoe.Me.Position;
			if (m.Name == "Mortality Experimenter" && Kole != null && Kole.IsActive && ExilePather.PathDistance(myPos, Kole.Position) < 50)
            {
                weight += 100;
            }	

            if (m.IsTargetingMe)
            {
                weight += 20;
            }

            if (m.Rarity == Rarity.Magic)
            {
                weight += 5;
            }
            else if (m.Rarity == Rarity.Rare)
            {
                weight += 10;
            }
            else if (m.Rarity == Rarity.Unique)
            {
                weight += 15;
            }

            // Minions that get in the way.
            switch (m.Name)
            {
                case "Summoned Skeleton":
                    weight -= 15;
                    break;

                case "Raised Zombie":
                    weight -= 15;
                    break;
            }

            if (m.Rarity == Rarity.Normal && m.Type.Contains("/Totems/"))
            {
                weight -= 15;
            }

            // Necros
            if (m.ExplicitAffixes.Any(a => a.InternalName.Contains("RaisesUndead")) ||
                m.ImplicitAffixes.Any(a => a.InternalName.Contains("RaisesUndead")))
            {
                weight += 30;
            }
        }

        private bool CombatTargetingOnInclusionCalcuation(NetworkObject entity)
        {
            try
            {
                var m = entity as Monster;
                if (m == null)
                    return false;

                if (AreaStateCache.Current.IsBlacklisted(m))
                    return false;

                // Do not consider inactive/dead mobs.
                if (!m.IsActive)
                    return false;

                // Ignore any mob that cannot die.
                if (m.CannotDie)
                    return false;

                // Ignore mobs that are too far to care about.
                if (m.Distance > (_currentLeashRange != -1 ? FlickerStrikeSettings.Instance.CombatRange : 300))
                    return false;
            }
            catch (Exception ex)
            {
                Log.Error("[CombatOnInclusionCalcuation]", ex);
                return false;
            }
            return true;
        }

        #endregion

        #region Implementation of IBase

        /// <summary>Initializes this routine.</summary>
        public void Initialize()
        {
            Log.DebugFormat("[FlickerStrike] Initialize");

            _combatTargeting.InclusionCalcuation += CombatTargetingOnInclusionCalcuation;
            _combatTargeting.WeightCalculation += CombatTargetingOnWeightCalculation;
			
        }

        #endregion

        #region Implementation of IDisposable

        /// <summary> </summary>
        public void Dispose()
        {
        }

        #endregion

        #region Implementation of IAuthored

        /// <summary>The name of the routine.</summary>
        public string Name
        {
            get { return "FlickerStrike"; }
        }

        /// <summary>The description of the routine.</summary>
        public string Description
        {
            get { return "A Flicker Strike routine for Exilebuddy."; }
        }

        /// <summary>
        /// The author of this object.
        /// </summary>
        public string Author
        {
            get { return "Naut"; }
        }

        /// <summary>
        /// The version of this routone.
        /// </summary>
        public Version Version
        {
            get { return new Version(0, 0, 1, 1); }
        }

        #endregion

        #region Implementation of IRunnable

        /// <summary> The routine start callback. Do any initialization here. </summary>
        public void Start()
        {
            Log.DebugFormat("[FlickerStrike] Start");

            _needsUpdate = true;

            GameEventManager.AreaChanged += GameEventManagerOnAreaChanged;
            GameEventManager.PlayerDied += GameEventManagerOnPlayerDied;

            if (FlickerStrikeSettings.Instance.FlickerStrikeSlot == -1 &&
                FlickerStrikeSettings.Instance.FrenzyChargeSlot == -1)
            {
                Log.ErrorFormat(
                    "[Start] Please configure the FlickerStrike settings (Settings -> FlickerStrike) before starting!");
                BotManager.Stop();
            }
        }

        private bool IsCastableHelper(Skill skill)
        {
            return skill != null && skill.IsCastable && !skill.IsTotem && !skill.IsTrap && !skill.IsMine;
        }

        /// <summary> The routine tick callback. Do any update logic here. </summary>
        public void Tick()
        {
            if (!LokiPoe.IsInGame)
                return;

            if (_needsUpdate)
            {
                _raiseZombieSlot = -1;
                _raiseSpectreSlot = -1;
                _animateWeaponSlot = -1;
                _animateGuardianSlot = -1;
                _flameblastSlot = -1;
                _enduringCrySlot = -1;
                _moltenShellSlot = -1;
                _arcticArmourSlot = -1;
                _vaalDiscSlot = -1;
                _auraSlot = -1;
                _totemSlot = -1;
                _trapSlot = -1;
                _coldSnapSlot = -1;
                _bloodRageSlot = -1;
                _rfSlot = -1;
                _summonSkeletonsSlot = -1;
                _summonRagingSpiritSlot = -1;
                _summonSkeletonCount = 0;
                _summonRagingSpiritCount = 0;
                _mineSlot = -1;
                _curseSlots.Clear();

                // Register curses.
                foreach (var skill in LokiPoe.InGameState.SkillBarPanel.Skills)
                {
                    var tags = skill.SkillTags;

                    if (tags.Contains("curse"))
                    {
                        var slot = skill.Slot;
                        if (slot != -1 && skill.IsCastable)
                        {
                            _curseSlots.Add(slot);
                        }
                    }

                    if (((skill.SkillTags.Contains("aura") && !tags.Contains("vaal")) || (skill.Name == "Tempest Shield") || (skill.Name == "Herald of Ice") || (skill.Name == "Herald of Thunder") || (skill.Name == "Herald of Ash")) && _auraSlot == -1)
                    {
                        _auraSlot = skill.Slot;
                    }
					
                    if (skill.IsTotem && _totemSlot == -1)
                    {
                        _totemSlot = skill.Slot;
                    }

                    if (skill.IsTrap && _trapSlot == -1)
                    {
                        _trapSlot = skill.Slot;
                    }

                    if (skill.IsMine && _mineSlot == -1)
                    {
                        _mineSlot = skill.Slot;
                    }
                }

                var cs = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Cold Snap");
                if (IsCastableHelper(cs))
                {
                    _coldSnapSlot = cs.Slot;
                }
				
				var vd = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Vaal Discipline");
                if (IsCastableHelper(vd))
                {
                    _vaalDiscSlot = vd.Slot;
                }

                var ss = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Summon Skeletons");
                if (IsCastableHelper(ss))
                {
                    _summonSkeletonsSlot = ss.Slot;
                }

                var srs = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Summon Raging Spirit");
                if (IsCastableHelper(srs))
                {
                    _summonRagingSpiritSlot = srs.Slot;
                }

                var rf = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Righteous Fire");
                if (IsCastableHelper(rf))
                {
                    _rfSlot = rf.Slot;
                }

                var br = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Blood Rage");
                if (IsCastableHelper(br))
                {
                    _bloodRageSlot = br.Slot;
                }

                var aa = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Arctic Armour");
                if (IsCastableHelper(aa))
                {
                    _arcticArmourSlot = aa.Slot;
                }

                var mc = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Molten Shell");
                if (IsCastableHelper(mc))
                {
                    _moltenShellSlot = mc.Slot;
                }

                var ec = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Enduring Cry");
                if (IsCastableHelper(ec))
                {
                    _enduringCrySlot = ec.Slot;
                }

                var rz = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Raise Zombie");
                if (IsCastableHelper(rz))
                {
                    _raiseZombieSlot = rz.Slot;
                }

                var rs = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Raise Spectre");
                if (IsCastableHelper(rs))
                {
                    _raiseSpectreSlot = rs.Slot;
                }

                var fb = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Flameblast");
                if (IsCastableHelper(fb))
                {
                    _flameblastSlot = fb.Slot;
                }

                var ag = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Animate Guardian");
                if (IsCastableHelper(ag))
                {
                    _animateGuardianSlot = ag.Slot;
                }

                var aw = LokiPoe.InGameState.SkillBarPanel.Skills.FirstOrDefault(s => s.Name == "Animate Weapon");
                if (IsCastableHelper(aw))
                {
                    _animateWeaponSlot = aw.Slot;
                }

                _needsUpdate = false;
            }
        }

        /// <summary> The routine stop callback. Do any pre-dispose cleanup here. </summary>
        public void Stop()
        {
            Log.DebugFormat("[FlickerStrike] Stop");

            GameEventManager.AreaChanged -= GameEventManagerOnAreaChanged;
        }

        #endregion

        #region Implementation of IConfigurable

        /// <summary> The bot's settings control. This will be added to the Exilebuddy Settings tab.</summary>
        public UserControl Control
        {
            get
            {
                using (var fs = new FileStream(@"Routines\FlickerStrike\SettingsGui.xaml", FileMode.Open))
                {
                    var root = (UserControl) XamlReader.Load(fs);

                    // Your settings binding here.

                    if (
                        !Wpf.SetupCheckBoxBinding(root, "AlwaysAttackInPlaceCheckBox", "AlwaysAttackInPlace",
                            BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupCheckBoxBinding failed for 'AlwaysAttackInPlaceCheckBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (
                        !Wpf.SetupComboBoxItemsBinding(root, "FlickerStrikeSlotComboBox", "AllSkillSlots",
                            BindingMode.OneWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupComboBoxItemsBinding failed for 'FlickerStrikeSlotComboBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (
                        !Wpf.SetupComboBoxSelectedItemBinding(root, "FlickerStrikeSlotComboBox",
                            "FlickerStrikeSlot", BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupComboBoxSelectedItemBinding failed for 'FlickerStrikeSlotComboBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (
                        !Wpf.SetupComboBoxItemsBinding(root, "FrenzyChargeSlotComboBox", "AllSkillSlots",
                            BindingMode.OneWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupComboBoxItemsBinding failed for 'FrenzyChargeSlotComboBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (
                        !Wpf.SetupComboBoxSelectedItemBinding(root, "FrenzyChargeSlotComboBox",
                            "FrenzyChargeSlot", BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupComboBoxSelectedItemBinding failed for 'FrenzyChargeSlotComboBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }


                    if (!Wpf.SetupTextBoxBinding(root, "CombatRangeTextBox", "CombatRange",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat("[SettingsControl] SetupTextBoxBinding failed for 'CombatRangeTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "MaxMeleeRangeTextBox", "MaxMeleeRange",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat("[SettingsControl] SetupTextBoxBinding failed for 'MaxMeleeRangeTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "MaxRangeRangeTextBox", "MaxRangeRange",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat("[SettingsControl] SetupTextBoxBinding failed for 'MaxRangeRangeTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "MaxFlameBlastChargesTextBox", "MaxFlameBlastCharges",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'MaxFlameBlastChargesTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "MoltenShellDelayMsTextBox", "MoltenShellDelayMs",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat("[SettingsControl] SetupTextBoxBinding failed for 'MoltenShellDelayMsTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "ArcticArmourCastModeTextBox", "ArcticArmourCastMode",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'ArcticArmourCastModeTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "TotemDelayMsTextBox", "TotemDelayMs",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'TotemDelayMsTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }
					
					if (!Wpf.SetupTextBoxBinding(root, "VaalDiscTriggerTextBox", "VaalDiscTrigger",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'VaalDiscTriggerTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "TrapDelayMsTextBox", "TrapDelayMs",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'TrapDelayMsTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (
                        !Wpf.SetupTextBoxBinding(root, "SummonRagingSpiritCountPerDelayTextBox",
                            "SummonRagingSpiritCountPerDelay",
                            BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'SummonRagingSpiritCountPerDelayTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "SummonRagingSpiritDelayMsTextBox", "SummonRagingSpiritDelayMs",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'SummonRagingSpiritDelayMsTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (
                        !Wpf.SetupTextBoxBinding(root, "SummonSkeletonCountPerDelayTextBox",
                            "SummonSkeletonCountPerDelay",
                            BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'SummonSkeletonCountPerDelayTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "SummonSkeletonDelayMsTextBox", "SummonSkeletonDelayMs",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'SummonSkeletonDelayMsTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    if (!Wpf.SetupTextBoxBinding(root, "MineDelayMsTextBox", "MineDelayMs",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'MineDelayMsTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }
					
					if (!Wpf.SetupTextBoxBinding(root, "FlickerStrikeCooldownMsTextBox", "FlickerStrikeCooldownMs",
                        BindingMode.TwoWay, FlickerStrikeSettings.Instance))
                    {
                        Log.DebugFormat(
                            "[SettingsControl] SetupTextBoxBinding failed for 'FlickerStrikeCooldownMsTextBox'.");
                        throw new Exception("The SettingsControl could not be created.");
                    }

                    // Your settings event handlers here.

                    return root;
                }
            }
        }

        /// <summary>The settings object. This will be registered in the current configuration.</summary>
        public JsonSettings Settings
        {
            get { return FlickerStrikeSettings.Instance; }
        }

        #endregion

        #region Implementation of IRoutine

        /// <summary>
        /// Sends data to the routine with the associated name.
        /// </summary>
        /// <param name="name">The name of the configuration.</param>
        /// <param name="param">The data passed for the configuration.</param>
        public void SetConfiguration(string name, params object[] param)
        {
            if (name == "leash")
            {
                _currentLeashRange = (int) param[0];
            }
        }

        /// <summary>
        /// Requests data from the routine with the associated name.
        /// </summary>
        /// <param name="name">The name of the configuration.</param>
        /// <returns>Data from the routine.</returns>
        public object GetConfiguration(string name)
        {
            return null;
        }

        private WorldItem BestAnimateGuardianTarget(Monster monster, int maxLevel)
        {
            var worldItems = LokiPoe.ObjectManager.GetObjectsByType<WorldItem>()
                .Where(wi => !_ignoreAnimatedItems.Contains(wi.Id) && wi.Distance < 30)
                .OrderBy(wi => wi.Distance);
            foreach (var wi in worldItems)
            {
                var item = wi.Item;
                if (item.RequiredLevel <= maxLevel &&
                    item.IsIdentified &&
                    !item.IsChromatic &&
                    item.SocketCount < 5 &&
                    item.MaxLinkCount < 5 &&
                    item.Rarity <= Rarity.Magic
                    )
                {
                    if (monster == null || monster.LeftHandWeaponVisual == "")
                    {
                        if (item.IsClawType || item.IsOneHandAxeType || item.IsOneHandMaceType ||
                            item.IsOneHandSwordType ||
                            item.IsOneHandThrustingSwordType || item.IsTwoHandAxeType || item.IsTwoHandMaceType ||
                            item.IsTwoHandSwordType)
                        {
                            _ignoreAnimatedItems.Add(wi.Id);
                            return wi;
                        }
                    }

                    if (monster == null || monster.ChestVisual == "")
                    {
                        if (item.IsBodyArmorType)
                        {
                            _ignoreAnimatedItems.Add(wi.Id);
                            return wi;
                        }
                    }

                    if (monster == null || monster.HelmVisual == "")
                    {
                        if (item.IsHelmetType)
                        {
                            _ignoreAnimatedItems.Add(wi.Id);
                            return wi;
                        }
                    }

                    if (monster == null || monster.BootsVisual == "")
                    {
                        if (item.IsBootType)
                        {
                            _ignoreAnimatedItems.Add(wi.Id);
                            return wi;
                        }
                    }

                    if (monster == null || monster.GlovesVisual == "")
                    {
                        if (item.IsGloveType)
                        {
                            _ignoreAnimatedItems.Add(wi.Id);
                            return wi;
                        }
                    }
                }
            }

            return null;
        }

        private WorldItem BestAnimateWeaponTarget(int maxLevel)
        {
            var worldItems = LokiPoe.ObjectManager.GetObjectsByType<WorldItem>()
                .Where(wi => !_ignoreAnimatedItems.Contains(wi.Id) && wi.Distance < 30)
                .OrderBy(wi => wi.Distance);
            foreach (var wi in worldItems)
            {
                var item = wi.Item;
                if (item.IsIdentified &&
                    item.RequiredLevel <= maxLevel &&
                    !item.IsChromatic &&
                    item.SocketCount < 5 &&
                    item.MaxLinkCount < 5 &&
                    item.Rarity <= Rarity.Magic &&
                    (item.IsClawType || item.IsOneHandAxeType || item.IsOneHandMaceType || item.IsOneHandSwordType ||
                     item.IsOneHandThrustingSwordType || item.IsTwoHandAxeType || item.IsTwoHandMaceType ||
                     item.IsTwoHandSwordType || item.IsDaggerType || item.IsStaffType))
                {
                    _ignoreAnimatedItems.Add(wi.Id);
                    return wi;
                }
            }
            return null;
        }

        private Monster BestDeadTarget
        {
            get
            {
                var myPos = LokiPoe.Me.Position;
                return LokiPoe.ObjectManager.GetObjectsByType<Monster>()
                    .Where(
                        m =>
                            m.Distance < 30 && m.IsActiveDead && m.Rarity != Rarity.Unique && m.CorpseUsable &&
                            ExilePather.PathDistance(myPos, m.Position) < 30)
                    .OrderBy(m => m.Distance).FirstOrDefault();
            }
        }

        private async Task<bool> CombatLogicEnd()
        {
            if (_arcticArmourSlot != -1 && FlickerStrikeSettings.Instance.ArcticArmourCastMode == 2)
            {
                if (LokiPoe.Me.HasArcticArmourBuff)
                {
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_arcticArmourSlot);

                    var err1 = LokiPoe.InGameState.SkillBarPanel.Use(_arcticArmourSlot, true);
                    if (err1 == LokiPoe.InGameState.UseError.None)
                    {
                        await Coroutine.Sleep(Utility.LatencySafeValue(500));

                        return true;
                    }

                    Log.ErrorFormat("[CombatLogicEnd] Use returned {0} for {1}.", err1, skill.Name);
                }
            }

            return false;
        }

        private async Task<bool> HandleShrines()
        {
            // TODO: Shrines need speical CR logic, because it's now the CRs responsibility for handling all combaat situations,
            // and shrines are now considered a combat situation due their nature.

            // Check for any active shrines.
            var shrines =
                LokiPoe.ObjectManager.Objects.OfType<Shrine>()
                    .Where(s => !s.IsDeactivated && s.Distance < 50)
                    .OrderBy(s => s.Distance)
                    .ToList();

            if (!shrines.Any())
                return false;
            
            // For now, just take the first shrine found.

            var shrine = shrines[0];

            // Handle Skeletal Shrine in a special way, or handle priority between multiple shrines at the same time.
            var skellyOverride = shrine.ShrineId == "Skeletons";

            // Try and only move to touch it when we have a somewhat navigable path.
            if ((Utility.NumberOfMobsBetween(LokiPoe.Me, shrine) < 5 &&
                Utility.NumberOfMobsNear(LokiPoe.Me, 20) < 3) || skellyOverride)
            {
                Log.DebugFormat("[Logic] Now moving towards the Shrine {0}.", shrine.Id);

                var myPos = LokiPoe.Me.Position;

                var pos = ExilePather.WalkablePositionFor(shrine, 10);

                // We need to filter out based on pathfinding, since otherwise, a large gap will lockup the bot.
                var pathDistance = ExilePather.PathDistance(myPos, pos);
                if (pathDistance > 50)
                {
                    return false;
                }

                if (ExilePather.CanObjectSee(LokiPoe.Me, pos) && myPos.Distance(pos) < 20)
                {
                    Log.DebugFormat("[Logic] Now attempting to interact with the Shrine {0}.", shrine.Id);

                    await Coroutines.FinishCurrentAction();

                    await Coroutines.InteractWith(shrine);

                    await Coroutine.Sleep(500);
                }
                else
                {
                    if (!PlayerMover.MoveTowards(pos))
                    {
                        Log.ErrorFormat("[HandleShrines] MoveTowards failed for {0}.", pos);
                    }
                }

                return true;
            }

            return false;
        }

        /// <summary>
        /// The routine's coroutine logic to execute.
        /// </summary>
        /// <param name="type">The requested type of logic to execute.</param>
        /// <returns></returns>
        // ReSharper disable once CSharpWarnings::CS1998
        public async Task<bool> Logic(string type)
        {
            if (type == "combat")
            {
                // Update targeting.
                CombatTargeting.Update();

                // This logic is now CR specific, because Strongbox gui labels interfere with targeting,
                // but not general movement using Move only.
                if (
                    LokiPoe.ObjectManager.GetObjectsByType<Chest>()
                        .Any(c => c.Distance < 70 && !c.IsOpened && c.IsStrongBox))
                {
                    if (LokiPoe.InGameState.IsAlwaysHighlightEnabled)
                    {
                        Log.InfoFormat("[Logic] Now disabling Always Highlight to avoid targeting issues.");
                        LokiPoe.Input.ToggleAlwaysHighlight();
                        return true;
                    }
                }
                else
                {
                    if (!LokiPoe.InGameState.IsAlwaysHighlightEnabled)
                    {
                        Log.InfoFormat("[Logic] Now enabling Always Highlight.");
                        LokiPoe.Input.ToggleAlwaysHighlight();
                        return true;
                    }
                }

                var myPos = LokiPoe.Me.Position;
				
				if (LokiPoe.Me.FrenzyCharges == 0)
				{
						_flickerstrikeStopwatch.Start();
				}
				if (LokiPoe.Me.FrenzyCharges == 1)
				{
						_flickerstrikeStopwatch.Restart();
						_flickerstrikeStopwatch.Stop();
				}

                // If we have flameblast, we need to use special logic for it.
                if (_flameblastSlot != -1)
                {
                    if (_castingFlameblast)
                    {
                        var c = LokiPoe.Me.FlameblastCharges;

                        // Handle being cast interrupted.
                        if (c < _lastFlameblastCharges)
                        {
                            _castingFlameblast = false;
                            return true;
                        }
                        _lastFlameblastCharges = c;

                        if (c >= FlickerStrikeSettings.Instance.MaxFlameBlastCharges)
                        {
                            // Stop using the skill, so it's cast.
                            await Coroutines.FinishCurrentAction();

                            _castingFlameblast = false;
                        }
                        else
                        {
                            // Keep using the skill to build up charges.
                            var buaerr = LokiPoe.InGameState.SkillBarPanel.UseAt(_flameblastSlot, false,
                                myPos);
                            if (buaerr != LokiPoe.InGameState.UseError.None)
                            {
                                Log.ErrorFormat("[Logic] UseAt returned {0} for {1}.", buaerr, "Flameblast");
                            }
                        }

                        return true;
                    }
                }

                // Limit this logic once per second, because it can get expensive and slow things down if run too fast.
                if (_animateGuardianSlot != -1 && _animateGuardianStopwatch.ElapsedMilliseconds > 1000)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_animateGuardianSlot);
                    if (skill.CanUse())
                    {
                        // Check for a target near us.
                        var target = BestAnimateGuardianTarget(skill.DeployedObjects.FirstOrDefault() as Monster,
                            skill.GetStat(StatType.AnimateItemMaximumLevelRequirement));
                        if (target != null)
                        {
                            Log.InfoFormat("[Logic] Using {0} on {1}.", skill.Name, target.Name);

                            var uaerr = LokiPoe.InGameState.SkillBarPanel.UseOn(_animateGuardianSlot, true, target);
                            if (uaerr == LokiPoe.InGameState.UseError.None)
                            {
                                await Coroutines.FinishCurrentAction(false);

                                await Coroutine.Sleep(Utility.LatencySafeValue(100));

                                // We need to remove the item highlighting.
                                LokiPoe.ProcessHookManager.ClearAllKeyStates();

                                return true;
                            }

                            Log.ErrorFormat("[Logic] UseOn returned {0} for {1}.", uaerr, skill.Name);
                        }

                        _animateGuardianStopwatch.Restart();
                    }
                }

                // Limit this logic once per second, because it can get expensive and slow things down if run too fast.
                if (_animateWeaponSlot != -1 && _animateWeaponStopwatch.ElapsedMilliseconds > 1000)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_animateWeaponSlot);
                    if (skill.CanUse())
                    {
                        // Check for a target near us.
                        var target = BestAnimateWeaponTarget(skill.GetStat(StatType.AnimateItemMaximumLevelRequirement));
                        if (target != null)
                        {
                            Log.InfoFormat("[Logic] Using {0} on {1}.", skill.Name, target.Name);

                            var uaerr = LokiPoe.InGameState.SkillBarPanel.UseOn(_animateWeaponSlot, true, target);
                            if (uaerr == LokiPoe.InGameState.UseError.None)
                            {
                                await Coroutines.FinishCurrentAction(false);

                                await Coroutine.Sleep(Utility.LatencySafeValue(100));

                                // We need to remove the item highlighting.
                                LokiPoe.ProcessHookManager.ClearAllKeyStates();

                                _animateWeaponStopwatch.Restart();

                                return true;
                            }

                            Log.ErrorFormat("[Logic] UseOn returned {0} for {1}.", uaerr, skill.Name);
                        }

                        _animateWeaponStopwatch.Restart();
                    }
                }

                // If we have Raise Spectre, we can look for dead bodies to use for our army as we move around.
                if (_raiseSpectreSlot != -1)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_raiseSpectreSlot);
                    if (skill.CanUse())
                    {
                        var max = skill.GetStat(StatType.NumberOfSpectresAllowed);
                        if (skill.NumberDeployed < max)
                        {
                            // Check for a target near us.
                            var target = BestDeadTarget;
                            if (target != null)
                            {
                                Log.InfoFormat("[Logic] Using {0} on {1}.", skill.Name, target.Name);

                                var uaerr = LokiPoe.InGameState.SkillBarPanel.UseAt(_raiseSpectreSlot, false,
                                    target.Position);
                                if (uaerr == LokiPoe.InGameState.UseError.None)
                                {
                                    await Coroutines.FinishCurrentAction(false);

                                    await Coroutine.Sleep(Utility.LatencySafeValue(100));

                                    return true;
                                }

                                Log.ErrorFormat("[Logic] UseAt returned {0} for {1}.", uaerr, skill.Name);
                            }
                        }
                    }
                }

                // If we have Raise Zombie, we can look for dead bodies to use for our army as we move around.
                if (_raiseZombieSlot != -1)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_raiseZombieSlot);
                    if (skill.CanUse())
                    {
                        var max = skill.GetStat(StatType.NumberOfZombiesAllowed);
                        if (skill.NumberDeployed < max)
                        {
                            // Check for a target near us.
                            var target = BestDeadTarget;
                            if (target != null)
                            {
                                Log.InfoFormat("[Logic] Using {0} on {1}.", skill.Name, target.Name);

                                var uaerr = LokiPoe.InGameState.SkillBarPanel.UseAt(_raiseZombieSlot, false,
                                    target.Position);
                                if (uaerr == LokiPoe.InGameState.UseError.None)
                                {
                                    await Coroutines.FinishCurrentAction(false);

                                    await Coroutine.Sleep(Utility.LatencySafeValue(100));

                                    return true;
                                }

                                Log.ErrorFormat("[Logic] UseAt returned {0} for {1}.", uaerr, skill.Name);
                            }
                        }
                    }
                }
				
				// Cast Vaal Discipline if we have it and energy shield falls below threshold.
                if (_vaalDiscSlot != -1)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_vaalDiscSlot);
                    if (skill.CanUse() && (LokiPoe.Me.EnergyShieldPercent < FlickerStrikeSettings.Instance.VaalDiscTrigger))
                    {
                        var err1 = LokiPoe.InGameState.SkillBarPanel.Use(_vaalDiscSlot, true);
                        if (err1 == LokiPoe.InGameState.UseError.None)
                        {
                            await Coroutine.Sleep(Utility.LatencySafeValue(500));

                            await Coroutines.FinishCurrentAction(false);

                            await Coroutine.Sleep(Utility.LatencySafeValue(100));

                            return true;
                        }

                        Log.ErrorFormat("[Logic] Use returned {0} for {1}.", err1, skill.Name);
                    }
                }

                // Since EC has a cooldown, we can just cast it when mobs are in range to keep our endurance charges refreshed.
                if (_enduringCrySlot != -1)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_enduringCrySlot);
                    if (skill.CanUse())
                    {
                        if (Utility.NumberOfMobsNear(LokiPoe.Me, 30) > 0)
                        {
                            var err1 = LokiPoe.InGameState.SkillBarPanel.Use(_enduringCrySlot, true);
                            if (err1 == LokiPoe.InGameState.UseError.None)
                            {
                                await Coroutine.Sleep(Utility.LatencySafeValue(500));

                                await Coroutines.FinishCurrentAction(false);

                                return true;
                            }

                            Log.ErrorFormat("[Logic] Use returned {0} for {1}.", err1, skill.Name);
                        }
                    }
                }

                // For Molten Shell, we want to limit cast time, since mobs that break the shield often would cause the CR to cast it over and over.
                if (_moltenShellSlot != -1 &&
                    _moltenShellStopwatch.ElapsedMilliseconds >= FlickerStrikeSettings.Instance.MoltenShellDelayMs)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_moltenShellSlot);
                    if (!LokiPoe.Me.HasMoltenShellBuff && skill.CanUse())
                    {
                        if (Utility.NumberOfMobsNear(LokiPoe.Me, FlickerStrikeSettings.Instance.CombatRange) > 0)
                        {
                            var err1 = LokiPoe.InGameState.SkillBarPanel.Use(_moltenShellSlot, true);

                            _moltenShellStopwatch.Restart();

                            if (err1 == LokiPoe.InGameState.UseError.None)
                            {
                                await Coroutine.Sleep(Utility.LatencySafeValue(500));

                                await Coroutines.FinishCurrentAction(false);

                                await Coroutine.Sleep(Utility.LatencySafeValue(100));

                                return true;
                            }

                            Log.ErrorFormat("[Logic] Use returned {0} for {1}.", err1, skill.Name);
                        }
                    }
                }

                // Handle Arctic Armour conditionally.
                if (_arcticArmourSlot != -1)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_arcticArmourSlot);
                    if (!LokiPoe.Me.HasArcticArmourBuff && skill.CanUse())
                    {
                        if (FlickerStrikeSettings.Instance.ArcticArmourCastMode == 1 ||
                            (FlickerStrikeSettings.Instance.ArcticArmourCastMode == 2 &&
                             Utility.NumberOfMobsNear(LokiPoe.Me, FlickerStrikeSettings.Instance.CombatRange) > 0))
                        {
                            var err1 = LokiPoe.InGameState.SkillBarPanel.Use(_arcticArmourSlot, true);
                            if (err1 == LokiPoe.InGameState.UseError.None)
                            {
                                await Coroutine.Sleep(Utility.LatencySafeValue(500));

                                await Coroutines.FinishCurrentAction(false);

                                await Coroutine.Sleep(Utility.LatencySafeValue(100));

                                return true;
                            }

                            Log.ErrorFormat("[Logic] Use returned {0} for {1}.", err1, skill.Name);
                        }
                    }
                }

                // Handle aura logic.
                if (_auraSlot != -1)
                {
                    foreach (var skill in LokiPoe.InGameState.SkillBarPanel.Skills)
                    {
                        var ts = false;
						var heraldice = false;
						var heraldash = false;
						var heraldthunder = false;

                        if (skill.Name == "Tempest Shield")
                            ts = true;
						
						if (skill.Name == "Herald of Ice")
                            heraldice = true;
						
						if (skill.Name == "Herald of Ash")
                            heraldash = true;	
							
						if (skill.Name == "Herald of Thunder")
                            heraldthunder = true;		

                        if (heraldthunder ||heraldash || heraldice || ts || (skill.SkillTags.Contains("aura") && !skill.SkillTags.Contains("vaal")))
                        {
                            if (skill.CanUse(false, true) && !LokiPoe.Me.HasAura(skill.Name) && (!ts || !LokiPoe.Me.HasTempestShieldBuff) && (!heraldice || !LokiPoe.Me.HasHeraldOfIceBuff) && (!heraldash || !LokiPoe.Me.HasHeraldOfAshBuff) && (!heraldthunder || !LokiPoe.Me.HasHeraldOfThunderBuff))
                            {
                                var doCast = true;

                                while (skill.Slot == -1)
                                {
                                    Log.InfoFormat("[Logic] Now assigning {0} to the skillbar.", skill.Name);

                                    var sserr = LokiPoe.InGameState.SkillBarPanel.SetSlot(_auraSlot, skill);

                                    if (sserr != LokiPoe.InGameState.SetSlotError.None)
                                    {
                                        Log.ErrorFormat("[Logic] SetSlot returned {0}.", sserr);

                                        doCast = false;

                                        break;
                                    }

                                    await Coroutine.Sleep(Utility.LatencySafeValue(1000));
                                }

                                if (!doCast)
                                {
                                    continue;
                                }

                                await Coroutines.FinishCurrentAction();

                                await Coroutine.Sleep(Utility.LatencySafeValue(1000));

                                var err1 = LokiPoe.InGameState.SkillBarPanel.Use(skill.Slot, false);
                                if (err1 == LokiPoe.InGameState.UseError.None)
                                {
                                    await Coroutine.Sleep(Utility.LatencySafeValue(250));

                                    await Coroutines.FinishCurrentAction(false);

                                    await Coroutine.Sleep(Utility.LatencySafeValue(1000));

                                    return true;
                                }
								
								var aura = LokiPoe.InGameState.SkillBarPanel.Slot(_auraSlot);
                                Log.ErrorFormat("[Logic] Use returned {0} for {1}.", err1, aura.Name);
                            }
                        }
                    }
                }

                // Check for a surround to use flameblast, just example logic.
                if (_flameblastSlot != -1)
                {
                    if (Utility.NumberOfMobsNear(LokiPoe.Me, 15) >= 4)
                    {
                        _castingFlameblast = true;
                        _lastFlameblastCharges = 0;
                        return true;
                    }
                }

                // TODO: _currentLeashRange of -1 means we need to use a cached location system to prevent back and forth issues of mobs despawning.

                // This is pretty important. Otherrwise, components can go invalid and exceptions are thrown.
                var bestTarget = CombatTargeting.Targets<Monster>().FirstOrDefault();

                // No monsters, we can execute non-critical combat logic, like buffs, auras, etc...
                // For this example, just going to continue executing bot logic.
                if (bestTarget == null)
                {
                    if (await HandleShrines())
                    {
                        return true;
                    }

                    return await CombatLogicEnd();
                }

                var cachedPosition = bestTarget.Position;
                var cachedId = bestTarget.Id;
                var cachedName = bestTarget.Name;
                var cachedRarity = bestTarget.Rarity;
                var cachedDistance = bestTarget.Distance;
                var cachedIsCursable = bestTarget.IsCursable;
                var cachedCurseCount = bestTarget.CurseCount;
                var cachedHasCurseFrom = new Dictionary<string, bool>();
                var cachedNumberOfMobsNear = Utility.NumberOfMobsNear(bestTarget, 15);
                var cachedProxShield = bestTarget.HasProximityShield;
                var _cachedMobsNearForAoe = Utility.NumberOfMobsNear(LokiPoe.Me,
                    FlickerStrikeSettings.Instance.MaxMeleeRange);
                var _cachedMobsNearForCurse = Utility.NumberOfMobsNear(bestTarget, 20);

                foreach (var curseSlot in _curseSlots)
                {
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(curseSlot);
                    cachedHasCurseFrom.Add(skill.Name, bestTarget.HasCurseFrom(skill.Name));
                }

                var canSee = ExilePather.CanObjectSee(LokiPoe.Me, bestTarget);
                var pathDistance = ExilePather.PathDistance(myPos, cachedPosition);
                var blockedByDoor = Utility.ClosedDoorBetween(LokiPoe.Me, bestTarget);

                if (await HandleShrines())
                {
                    return true;
                }

                if (pathDistance.CompareTo(float.MaxValue) == 0)
                {
                    Log.ErrorFormat(
                        "[Logic] Could not determine the path distance to the best target. Now blacklisting it.");
                    AreaStateCache.Current.Blacklist(cachedId, TimeSpan.FromMinutes(1), "Unable to pathfind to.");
                    return true;
                }

                // Prevent combat loops from happening.
                if (pathDistance > FlickerStrikeSettings.Instance.CombatRange)
                    return false;

                if (!canSee || blockedByDoor)
                {
                    Log.InfoFormat(
                        "[Logic] Now moving towards the monster {0} because [canSee: {1}][pathDistance: {2}][blockedByDoor: {3}]",
                        cachedName, canSee, pathDistance, blockedByDoor);

                    if (!PlayerMover.MoveTowards(cachedPosition))
                    {
                        Log.ErrorFormat("[Logic] MoveTowards failed for {0}.", cachedPosition);
                    }

                    return true;
                }

                // Handle totem logic.
                if (_totemSlot != -1 &&
                    _totemStopwatch.ElapsedMilliseconds > FlickerStrikeSettings.Instance.TotemDelayMs)
                {
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_totemSlot);
                    if (skill.CanUse() &&
                        !skill.DeployedObjects.Select(o => o as Monster).Any(t => !t.IsDead && t.Distance < 60))
                    {
                        var err1 = LokiPoe.InGameState.SkillBarPanel.UseAt(_totemSlot, true,
                            Utility.CalculatePointAtDistanceAfterStart(myPos, cachedPosition,
                                cachedDistance/2));

                        _totemStopwatch.Restart();

                        if (err1 == LokiPoe.InGameState.UseError.None)
                            return true;

                        Log.ErrorFormat("[Logic] UseAt returned {0} for {1}.", err1, skill.Name);
                    }
                }

                // Handle trap logic.
                if (_trapSlot != -1 &&
                    _trapStopwatch.ElapsedMilliseconds > FlickerStrikeSettings.Instance.TrapDelayMs)
                {
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_trapSlot);
                    if (skill.CanUse())
                    {
                        var err1 = LokiPoe.InGameState.SkillBarPanel.UseAt(_trapSlot, true,
                            Utility.CalculatePointAtDistanceAfterStart(myPos, cachedPosition,
                                cachedDistance/2));

                        _trapStopwatch.Restart();

                        if (err1 == LokiPoe.InGameState.UseError.None)
                            return true;

                        Log.ErrorFormat("[Logic] UseAt returned {0} for {1}.", err1, skill.Name);
                    }
                }

                // Handle curse logic - curse magic+ and packs of 4+, but only cast within MaxRangeRange.
                var checkCurses = myPos.Distance(cachedPosition) < FlickerStrikeSettings.Instance.MaxRangeRange &&
                                  (cachedRarity >= Rarity.Magic || _cachedMobsNearForCurse >= 1);
                if (checkCurses)
                {
                    foreach (var curseSlot in _curseSlots)
                    {
                        var skill = LokiPoe.InGameState.SkillBarPanel.Slot(curseSlot);
                        if (skill.CanUse() &&
                            cachedIsCursable &&
                            cachedCurseCount < LokiPoe.Me.TotalCursesAllowed &&
                            !cachedHasCurseFrom[skill.Name])
                        {
                            var err1 = LokiPoe.InGameState.SkillBarPanel.UseAt(curseSlot, true, cachedPosition);
                            if (err1 == LokiPoe.InGameState.UseError.None)
                                return true;

                            Log.ErrorFormat("[Logic] Use returned {0} for {1}.", err1, skill.Name);
                        }
                    }
                }

                // Simply cast Blood Rage if we have it.
                if (_bloodRageSlot != -1)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_bloodRageSlot);
                    if (skill.CanUse() && !LokiPoe.Me.HasBloodRageBuff)
                    {
                        var err1 = LokiPoe.InGameState.SkillBarPanel.Use(_bloodRageSlot, true);
                        if (err1 == LokiPoe.InGameState.UseError.None)
                        {
                            await Coroutine.Sleep(Utility.LatencySafeValue(500));

                            await Coroutines.FinishCurrentAction(false);

                            await Coroutine.Sleep(Utility.LatencySafeValue(100));

                            return true;
                        }

                        Log.ErrorFormat("[Logic] Use returned {0} for {1}.", err1, skill.Name);
                    }
                }

                // Simply cast RF if we have it.
                if (_rfSlot != -1)
                {
                    // See if we can use the skill.
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_rfSlot);
                    if (skill.CanUse() && !LokiPoe.Me.HasRighteousFireBuff)
                    {
                        var err1 = LokiPoe.InGameState.SkillBarPanel.Use(_rfSlot, true);
                        if (err1 == LokiPoe.InGameState.UseError.None)
                        {
                            await Coroutine.Sleep(Utility.LatencySafeValue(500));

                            await Coroutines.FinishCurrentAction(false);

                            await Coroutine.Sleep(Utility.LatencySafeValue(100));

                            return true;
                        }

                        Log.ErrorFormat("[Logic] Use returned {0} for {1}.", err1, skill.Name);
                    }
                }

                if (_summonRagingSpiritSlot != -1 &&
                    _summonRagingSpiritStopwatch.ElapsedMilliseconds >
                    FlickerStrikeSettings.Instance.SummonRagingSpiritDelayMs)
                {
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_summonRagingSpiritSlot);
                    var max = skill.GetStat(StatType.NumberOfRagingSpiritsAllowed);
                    if (skill.NumberDeployed < max && skill.CanUse())
                    {
                        ++_summonRagingSpiritCount;

                        var err1 = LokiPoe.InGameState.SkillBarPanel.Use(_summonRagingSpiritSlot, false);

                        if (_summonRagingSpiritCount >=
                            FlickerStrikeSettings.Instance.SummonRagingSpiritCountPerDelay)
                        {
                            _summonRagingSpiritCount = 0;
                            _summonRagingSpiritStopwatch.Restart();
                        }

                        if (err1 == LokiPoe.InGameState.UseError.None)
                        {
                            await Coroutine.Sleep(Utility.LatencySafeValue(500));

                            await Coroutines.FinishCurrentAction(false);

                            return true;
                        }

                        Log.ErrorFormat("[Logic] Use returned {0} for {1}.", err1, skill.Name);
                    }
                }

                if (_summonSkeletonsSlot != -1 &&
                    _summonSkeletonsStopwatch.ElapsedMilliseconds >
                    FlickerStrikeSettings.Instance.SummonSkeletonDelayMs)
                {
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_summonSkeletonsSlot);
                    var max = skill.GetStat(StatType.NumberOfSkeletonsAllowed);
                    if (skill.NumberDeployed < max && skill.CanUse())
                    {
                        ++_summonSkeletonCount;

                        var err1 = LokiPoe.InGameState.SkillBarPanel.UseAt(_summonSkeletonsSlot, true,
                            Utility.CalculatePointAtDistanceAfterStart(myPos, cachedPosition,
                                cachedDistance/2));

                        if (_summonSkeletonCount >= FlickerStrikeSettings.Instance.SummonSkeletonCountPerDelay)
                        {
                            _summonSkeletonCount = 0;
                            _summonSkeletonsStopwatch.Restart();
                        }

                        if (err1 == LokiPoe.InGameState.UseError.None)
                        {
                            await Coroutines.FinishCurrentAction(false);

                            return true;
                        }

                        Log.ErrorFormat("[Logic] UseAt returned {0} for {1}.", err1, skill.Name);
                    }
                }

                if (_mineSlot != -1 && _mineStopwatch.ElapsedMilliseconds >
                    FlickerStrikeSettings.Instance.MineDelayMs &&
                    myPos.Distance(cachedPosition) < FlickerStrikeSettings.Instance.MaxMeleeRange)
                {
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_mineSlot);
                    var max = skill.GetStat(StatType.SkillDisplayNumberOfRemoteMinesAllowed);
                    var insta = skill.GetStat(StatType.MineDetonationIsInstant) == 1;
                    if (skill.NumberDeployed < max && skill.CanUse())
                    {
                        var err1 = LokiPoe.InGameState.SkillBarPanel.Use(_mineSlot, true);

                        if (err1 == LokiPoe.InGameState.UseError.None)
                        {
                            await Coroutine.Sleep(Utility.LatencySafeValue(500));

                            await Coroutines.FinishCurrentAction(false);

                            if (!insta)
                            {
                                await Coroutine.Sleep(Utility.LatencySafeValue(350));

                                LokiPoe.Input.DetonateMines();
                            }

                            _mineStopwatch.Restart();

                            return true;
                        }

                        _mineStopwatch.Restart();

                        Log.ErrorFormat("[Logic] UseAt returned {0} for {1}.", err1, skill.Name);
                    }
                }

                // Handle cold snap logic. Only use when power charges won't be consumed.
                if (_coldSnapSlot != -1)
                {
                    var skill = LokiPoe.InGameState.SkillBarPanel.Slot(_coldSnapSlot);
                    if (skill.CanUse(false, false, false))
                    {
                        var err1 = LokiPoe.InGameState.SkillBarPanel.UseAt(_coldSnapSlot, true, cachedPosition);

                        if (err1 == LokiPoe.InGameState.UseError.None)
                            return true;

                        Log.ErrorFormat("[Logic] UseAt returned {0} for {1}.", err1, skill.Name);
                    }
                }

                var aip = false;

                var aoe = false;
                var melee = false;

                int slot;

                // Logic for figuring out if we should use an AoE skill or single target.
                if (cachedNumberOfMobsNear > 2 && cachedRarity < Rarity.Rare)
                {
                    aoe = true;
                }

                // Logic for figuring out if we should use an AoE skill instead.
                if (myPos.Distance(cachedPosition) < FlickerStrikeSettings.Instance.MaxMeleeRange)
                {
                    melee = true;
                    if (_cachedMobsNearForAoe >= 3)
                    {
                        aoe = true;
                    }
                    else
                    {
                        aoe = false;
                    }
                }

                // This sillyness is for making sure we always use a skill, and is why generic code is a PITA
                // when it can be configured like so.
                
                    if (_flickerstrikeStopwatch.ElapsedMilliseconds > FlickerStrikeSettings.Instance.FlickerStrikeCooldownMs || LokiPoe.Me.FrenzyCharges >= 1)
                    {
                        slot = EnsurceCast(FlickerStrikeSettings.Instance.FlickerStrikeSlot);
						melee = false;
                        if (slot == -1)
                        {
                            slot = EnsurceCast(FlickerStrikeSettings.Instance.FrenzyChargeSlot);
							melee = true;
                        }
                    }
                    else
                    {
                        slot = EnsurceCast(FlickerStrikeSettings.Instance.FrenzyChargeSlot);
                        melee = true;
                    }
                

                if (FlickerStrikeSettings.Instance.AlwaysAttackInPlace)
                    aip = true;

                if (slot == -1)
                {
                    slot = FlickerStrikeSettings.Instance.FlickerStrikeSlot;
                    melee = true;
                }

                if (melee || cachedProxShield)
                {
                    if (LokiPoe.Me.Position.Distance(cachedPosition) > FlickerStrikeSettings.Instance.MaxMeleeRange)
                    {
                        if (!PlayerMover.MoveTowards(cachedPosition))
                        {
                            Log.ErrorFormat("[Logic] MoveTowards failed for {0}.", cachedPosition);
                        }
                        return true;
                    }
                }
                else
                {
                    if (LokiPoe.Me.Position.Distance(cachedPosition) > FlickerStrikeSettings.Instance.MaxRangeRange)
                    {
                        if (!PlayerMover.MoveTowards(cachedPosition))
                        {
                            Log.ErrorFormat("[Logic] MoveTowards failed for {0}.", cachedPosition);
                        }
                        return true;
                    }
                }

                var err = LokiPoe.InGameState.SkillBarPanel.UseAt(slot, aip, cachedPosition);
                if (err != LokiPoe.InGameState.UseError.None)
                {
                    Log.ErrorFormat("[Logic] UseAt returned {0}.", err);
                }

                return true;
            }

            return false;
        }

        #endregion

        private int EnsurceCast(int slot)
        {
            if (slot == -1)
                return slot;

            var slotSkill = LokiPoe.InGameState.SkillBarPanel.Slot(slot);
            if (slotSkill == null || !slotSkill.CanUse())
            {
                return -1;
            }

            return slot;
        }

        #region Override of Object

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return Name + ": " + Description;
        }

        #endregion

        private void GameEventManagerOnAreaChanged(object sender, AreaChangedEventArgs areaChangedEventArgs)
        {
            _ignoreAnimatedItems.Clear();
			if (BasicGrindBotSettings.Instance.GrindZoneName == "The Lunaris Temple Level 2")
			{
				FlickerStrikeSettings.Instance.CombatRange = 1;
			}
			if (BasicGrindBotSettings.Instance.GrindZoneName != "The Lunaris Temple Level 2")
			{
				FlickerStrikeSettings.Instance.CombatRange = 60;
			}	
        }

        private void GameEventManagerOnPlayerDied(object sender, PlayerDiedEventArgs playerDiedEventArgs)
        {
        }
    }
}
